from sortedcontainers import SortedKeyList
import random 

class GuidedExchangeOperator:
    def __init__(self, environment_energies, feature_key):
        self.n_envs = int(len(environment_energies)/2)
        self.env_energy_differences = [environment_energies[i] - environment_energies[i + self.n_envs] for i in range(self.n_envs)]

        self.neigh_energy_difference = [environment_energies[i+1] - environment_energies[i] for i in range((self.n_envs*2)-1)] + [-1000]

        self.feature_key = feature_key

        self.symbol1_exchange_energies = dict()
        self.symbol2_exchange_energies = dict()

        self.symbol1_indices = SortedKeyList(key=lambda x: self.symbol1_exchange_energies[x])
        self.symbol2_indices = SortedKeyList(key=lambda x: self.symbol2_exchange_energies[x])

        self.n_symbol1_atoms = 0
        self.n_symbol2_atoms = 0

    def env_from_feature(self, x):
        return x % self.n_envs

    def guided_exchange(self, particle):
        symbol1_index = self.symbol1_indices[0]
        symbol2_index = self.symbol2_indices[0]

        particle.swap_symbols([(symbol1_index, symbol2_index)])
        return symbol1_index, symbol2_index

    def basin_hop_step(self, particle):
        expected_energy_gain = -1
        index = -1
        random_indices = random.sample(range(max(self.n_symbol1_atoms, self.n_symbol2_atoms)), max(self.n_symbol1_atoms, self.n_symbol2_atoms))
        
        while expected_energy_gain <= 0 and index < min(self.n_symbol1_atoms, self.n_symbol2_atoms):
            index += 1
            symbol1_index = self.symbol1_indices[random_indices[index] % self.n_symbol1_atoms]
            symbol1_energy = self.symbol1_exchange_energies[symbol1_index]

            symbol2_index = self.symbol2_indices[random_indices[index] % self.n_symbol2_atoms]
            symbol2_energy = self.symbol2_exchange_energies[symbol2_index]

            expected_energy_gain = symbol1_energy + symbol2_energy
            if expected_energy_gain > 0:
                particle.swap_symbols([(symbol1_index, symbol2_index)])
                return symbol1_index, symbol2_index

        symbol1_index = self.symbol1_indices[random_indices[index] % self.n_symbol1_atoms]
        symbol2_index = self.symbol2_indices[random_indices[index] % self.n_symbol2_atoms]

        particle.swap_symbols([(symbol1_index, symbol2_index)])
        return symbol1_index, symbol2_index

    def bind_particle(self, particle):
        symbols = sorted(particle.atoms.get_all_symbols())
        symbol1 = symbols[0]
        symbol2 = symbols[1]

        symbol1_indices = particle.get_indices_by_symbol(symbol1)
        symbol2_indices = particle.get_indices_by_symbol(symbol2)

        self.n_symbol1_atoms = len(symbol1_indices)
        self.n_symbol2_atoms = len(symbol2_indices)

        atom_features = particle.get_atom_features(self.feature_key)
        for index in symbol1_indices:
            feature = atom_features[index]
            
            energy_change = 0
            for neigh_idx in particle.get_coordination_atoms(index):
                neigh_feature = atom_features[neigh_idx]
                energy_change -= self.neigh_energy_difference[neigh_feature-1] 

            self.symbol1_exchange_energies[index] = (-self.env_energy_differences[self.env_from_feature(feature)] + energy_change)
            self.symbol1_indices.add(index)

        for index in symbol2_indices:
            feature = atom_features[index]
            
            energy_change = 0
            for neigh_idx in particle.get_coordination_atoms(index):
                neigh_feature = atom_features[neigh_idx]
                energy_change += self.neigh_energy_difference[neigh_feature] 

            self.symbol2_exchange_energies[index] = (+self.env_energy_differences[self.env_from_feature(feature)] + energy_change)
            self.symbol2_indices.add(index)         

    def update(self, particle, indices, exchange_indices):
        symbols = sorted(particle.atoms.get_all_symbols())
        symbol1 = symbols[0]

        atom_features = particle.get_atom_features(self.feature_key)
        for index in indices:
            if index in exchange_indices:
                if particle.get_symbol(index) == symbol1:
                    self.symbol2_indices.remove(index)
                else:
                    self.symbol1_indices.remove(index)
            else:
                if particle.get_symbol(index) == symbol1:
                    self.symbol1_indices.remove(index)
                else:
                    self.symbol2_indices.remove(index)

        for index in indices:
            feature = atom_features[index]
            new_exchange_energy = self.env_energy_differences[self.env_from_feature(feature)]
            if index in exchange_indices:
                if particle.get_symbol(index) == symbol1:
                    
                    energy_change = 0
                    for neigh_idx in particle.get_coordination_atoms(index):
                        neigh_feature = atom_features[neigh_idx]
                        energy_change -= self.neigh_energy_difference[neigh_feature-1]

                    self.symbol1_exchange_energies[index] = -new_exchange_energy + energy_change
                    del self.symbol2_exchange_energies[index]
                else:
                    energy_change = 0
                    for neigh_idx in particle.get_coordination_atoms(index):
                        neigh_feature = atom_features[neigh_idx]
                        energy_change += self.neigh_energy_difference[neigh_feature] 

                    self.symbol2_exchange_energies[index] = new_exchange_energy + energy_change
                    del self.symbol1_exchange_energies[index]
            else:
                if particle.get_symbol(index) == symbol1:

                    energy_change = 0
                    for neigh_idx in particle.get_coordination_atoms(index):
                        neigh_feature = atom_features[neigh_idx]
                        energy_change -= self.neigh_energy_difference[neigh_feature-1]

                    self.symbol1_exchange_energies[index] = -new_exchange_energy + energy_change
                else:
                    energy_change = 0
                    for neigh_idx in particle.get_coordination_atoms(index):
                        neigh_feature = atom_features[neigh_idx]
                        energy_change += self.neigh_energy_difference[neigh_feature] 
                    self.symbol2_exchange_energies[index] = new_exchange_energy + energy_change

        for index in indices:
            if particle.get_symbol(index) == symbol1:
                self.symbol1_indices.add(index)
            else:
                self.symbol2_indices.add(index)

    # def update(self, particle, indices, exchange_indices):
    #     symbols = sorted(particle.atoms.get_all_symbols())
    #     symbol1 = symbols[0]

    #     atom_features = particle.get_atom_features(self.feature_key)
    #     for index in indices:
    #         if index in exchange_indices:
    #             if particle.get_symbol(index) == symbol1:
    #                 self.symbol2_indices.remove(index)
    #                 #del self.symbol2_exchange_energies[index]
    #             else:
    #                 self.symbol1_indices.remove(index)
    #                 #del self.symbol1_exchange_energies[index]
    #         else:
    #             if particle.get_symbol(index) == symbol1:
    #                 self.symbol1_indices.remove(index)
    #             else:
    #                 self.symbol2_indices.remove(index)

    #     for index in indices:
    #         feature = atom_features[index]
    #         new_exchange_energy = self.env_energy_differences[self.env_from_feature(feature)]
    #         if particle.get_symbol(index) == symbol1:
    #             energy_change = 0
    #             for neigh_idx in particle.get_coordination_atoms(index):
    #                 neigh_feature = atom_features[neigh_idx]
    #                 energy_change -= self.neigh_energy_difference[neigh_feature-1]
    #             #self.symbol1_exchange_energies[index] = (-self.env_energy_differences[self.env_from_feature(feature)] + energy_change)
    #             #elf.symbol1_indices.add(index)
    #         else:
    #             energy_change = 0
    #             for neigh_idx in particle.get_coordination_atoms(index):
    #                 neigh_feature = atom_features[neigh_idx]
    #                 energy_change += self.neigh_energy_difference[neigh_feature+1]
    #             #self.symbol2_exchange_energies[index] = (+self.env_energy_differences[self.env_from_feature(feature)] + energy_change)
    #             #self.symbol2_indices.add(index)
            
    #         if index in exchange_indices:
    #             if particle.get_symbol(index) == symbol1:
    #                 self.symbol1_exchange_energies[index] = -new_exchange_energy + energy_change
    #                 del self.symbol2_exchange_energies[index]
    #             else:
    #                 self.symbol2_exchange_energies[index] = new_exchange_energy + energy_change
    #                 del self.symbol1_exchange_energies[index]
    #         else:
    #             if particle.get_symbol(index) == symbol1:
    #                 self.symbol1_exchange_energies[index] = -new_exchange_energy + energy_change
    #             else:
    #                 self.symbol2_exchange_energies[index] = new_exchange_energy + energy_change

    #     for index in indices:
    #         if particle.get_symbol(index) == symbol1:
    #             self.symbol1_indices.add(index)
    #         else:
    #             self.symbol2_indices.add(index)


from collections import defaultdict

import numpy as np
import itertools
import copy

from ase.neighborlist import natural_cutoffs
from ase.neighborlist import build_neighbor_list


class NeighborList:
    def __init__(self):
        self.list = defaultdict(lambda: set())

    def __getitem__(self, item):
        return self.list[item]

    def __setitem__(self, key, value):
        self.list[key] = value

    def construct(self, atoms, scale_factor=1.0, npl = True, cutoffs = None):

        if not cutoffs:
            cutoffs = natural_cutoffs(atoms, mult=scale_factor)
            
        neighbor_list = build_neighbor_list(atoms,
                                            cutoffs=cutoffs,
                                            bothways=True,
                                            self_interaction=False)

        for atom_idx, _ in enumerate(atoms):
            neighbors, _ = neighbor_list.get_neighbors(atom_idx)
            if npl:
                self.list[atom_idx] = set(neighbors)
            else:
                self.list[atom_idx] = neighbors

    def get_coordination_number(self, atom_idx):
        return len(self.list[atom_idx])

    def get_n_bonds(self):
        n_bonds = sum([len(l) for l in list(self.list.values())])
        return n_bonds/2

    def get_coordination_atoms(self, atom_idx):
        return list(self.list[atom_idx])
    
    def get_max_coordination_number(self, indices):
        if len(indices) == 1:
            return 12
        if len(indices) == 2:
            return 18
        if len(indices) == 3:
            return 22
        if len(indices) == 4:
            return 26

        common_atom_indices = []
        for pair in itertools.combinations(indices, 2):
            shared_pair = list(np.intersect1d(self.get_coordination_atoms(pair[0]), self.get_coordination_atoms(pair[1])))
            common_atom_indices += shared_pair    

        return 12*len(indices) - len(set(common_atom_indices))
    
    def get_generalized_coordination_number(self, indices):
        max_coordination = self.get_max_coordination_number(indices)
        coordination_atoms = []
        for atom_idx in indices:
            for neigh in self.get_coordination_atoms(atom_idx):
                if neigh not in coordination_atoms and neigh not in indices:
                    coordination_atoms.append(neigh)
       
        tot_cns = 0
        for neighbor_idx in coordination_atoms:
            cn = self.get_coordination_number(neighbor_idx)
            tot_cns += cn
        
        return round(tot_cns / max_coordination, 2)
    
    def get_atoms_in_the_surface_plane(self, atom_idx, edges_corner=False):
        """Find atoms within the same surface, excluding bulk atoms,
        atom must be in a terrace position. To include edges and cornes edges_corners=True"""
        cn = self.get_coordination_number(atom_idx)
        if cn != 9 and cn != 8:
            raise Exception('AtomIsNotInTerrace')

        atoms_indices_in_plane = [atom_idx]
        for atom_idx in atoms_indices_in_plane:
            neighbors = self.get_coordination_atoms(atom_idx)
            for neighbor in neighbors:
                cn_neighbor = self.get_coordination_number(neighbor)
                if cn_neighbor == cn and neighbor not in atoms_indices_in_plane:
                    atoms_indices_in_plane.append(neighbor)
        
        if edges_corner == True:
            for atom_idx in atoms_indices_in_plane:
                neighbors = self.get_coordination_atoms(atom_idx) 
                for neighbor in neighbors:
                    cn_neighbor = self.get_coordination_number(neighbor)
                    if cn_neighbor < cn and neighbor not in atoms_indices_in_plane:
                        atoms_indices_in_plane.append(neighbor)

        return atoms_indices_in_plane
        







from setuptools import setup, find_packages

setup(
    name="npl",
    version="1.0.0",
    description="Nanoparticle Library for computational analysis of nanoparticles",
    long_description="This is a long description for the Nanoparticle Library.",
    long_description_content_type="text/markdown",
    author="Riccardo Farris",
    author_email="rfarris@ub.edu",
    url="",
    packages=find_packages(),
    install_requires=[
        "acat>=1.7.1",
        "scikit_learn>=1.5.0",
        "scipy>=1.10.0",
        "sortedcontainers>=2.4.0"
    ],
)


import numpy as np
from collections import defaultdict

from ase import Atoms

# TODO Derive directly from Atoms


class AtomWrapper:
    """Wrapper class around the ase Atoms object to extend it with some higher-level functionality.

    This class extends an Atoms object with additional functions for ordering, stoichiometry
    handling and is split from BaseNanoparticle to keep the size of the modules as small as possible.
    """
    def __init__(self):
        self.atoms = Atoms()

        return

    def add_atoms(self, new_atoms):
        self.atoms.extend(new_atoms)

    def get_atoms(self, indices):
        return self.atoms[indices]

    def get_position(self,idx):
        return self.atoms[idx].position

    def remove_atoms(self, indices):
        del self.atoms[indices]

    def get_ase_atoms(self, indices=None):
        """Return atoms specified by the indices.

        Parameters:
            indices: list/array of int
        """
        if indices is None:
            return self.atoms
        return self.atoms[indices]

    def swap_symbols(self, index_pairs):
        for idx1, idx2 in index_pairs:
            self.atoms[idx1].symbol, self.atoms[idx2].symbol = self.atoms[idx2].symbol, self.atoms[idx1].symbol

    def random_ordering(self, new_stoichiometry):
        """Creates a random chemical ordering for the given stoichiometry

        The new stoichiometry has to be given as dict of type str -> int that matches
        the symbol to the number of atoms.

        Parameters:
            new_stoichiometry: dict
        """
        new_symbols = []
        for symbol in new_stoichiometry:
            new_symbols += [symbol]*new_stoichiometry[symbol]
        np.random.shuffle(new_symbols)
        self.atoms.symbols = new_symbols

    def transform_atoms(self, atom_indices, new_symbols):
        for idx, symbol in zip(atom_indices, new_symbols):
            self.atoms[idx].symbol = symbol

    def get_indices(self):
        """Convenience function for range(n_atoms)."""
        return np.arange(0, len(self.atoms))

    def get_all_symbols(self):
        """Return list of symbols which occur at least once in the particle."""
        return list(self.atoms.symbols.species())

    def get_symbol(self, atom_idx):
        return self.atoms[atom_idx].symbol

    def get_symbols(self, indices=None):
        if indices is None:
            return self.atoms.symbols
        return self.atoms[indices].symbols

    def get_indices_by_symbol(self, symbol):
        """Return atom indices of given symbol.

        Parameters:
            symbol : str
        """
        if symbol in self.atoms.symbols.indices():
            return self.atoms.symbols.indices()[symbol]
        else:
            return np.array([])

    def get_n_atoms(self):
        """Return the number of atoms."""
        return len(self.atoms)

    # TODO implement more efficiently using in-built ase methods
    def get_n_atoms_of_symbol(self, symbol):
        return len(self.get_indices_by_symbol(symbol))

    def get_stoichiometry(self):
        """Return the current composition as defaultdict.

        For symbols that are not in the particle, the defaultdict will evaluate to 0.
        """
        stoichiometry = self.atoms.symbols.indices()
        for symbol in stoichiometry:
            stoichiometry[symbol] = len(stoichiometry[symbol])
        return defaultdict(lambda: 0, stoichiometry)

    def get_positions(self, indices=None):
        """Return the positions of the atoms.

        If indices is not specified, all atoms will be returned.

        Parameters:
            indices : list/array of int
        """
        if indices is None:
            return self.atoms.positions
        return self.atoms[indices].positions

    def translate_atoms_positions(self, position):
        self.atoms.translate(-position)
        



# npl/core/__init__.py

from npl.core.base_nanoparticle import BaseNanoparticle
from npl.core.nanoparticle import Nanoparticle
from npl.core.neighbor_list import NeighborList
from npl.core.adsorption import FindAdsorptionSites

__all__ = [
    "BaseNanoparticle",
    "Nanoparticle",
    "NeighborList",
    "FindAdsorptionSites"
]


from .energy_calculator import (EMTCalculator,
                                BayesianRRCalculator,
                                EnergyCalculator,
                                compute_coefficients_for_linear_topological_model)

__all__ = [
    "EMTCalculator",
    "BayesianRRCalculator",
    "compute_coefficients_for_linear_topological_model"
]


import numpy as np

from npl.core.base_nanoparticle import BaseNanoparticle
from npl.core.adsorption import FindAdsorptionSites

from ase.cluster.octahedron import Octahedron
from ase import Atoms


class Nanoparticle(BaseNanoparticle):
    def __init__(self):
        BaseNanoparticle.__init__(self)
        self.adsorption = FindAdsorptionSites()

    def truncated_octahedron(self, height, cutoff, stoichiometry, lattice_constant=3.9, alloy=False):
        octa = Octahedron('Pt', height, cutoff, latticeconstant=lattice_constant, alloy=alloy)
        atoms = Atoms(octa.symbols, octa.positions)
        com = atoms.get_center_of_mass()
        atoms.positions -= com

        self.add_atoms(atoms, recompute_neighbor_list=False)
        self.random_ordering(stoichiometry)
        self.construct_neighbor_list()

    def adjust_stoichiometry(self, target_stoichiometry):
        def choose_n_atoms_with_symbol(symbol, n_atoms):
            atoms_with_symbol = self.get_indices_by_symbol(symbol)
            selected_atoms = np.random.choice(atoms_with_symbol, n_atoms, replace=False)
            return selected_atoms

        excess_atoms = np.array([])
        for symbol in self.get_stoichiometry():
            if symbol in target_stoichiometry:
                difference = self.get_stoichiometry()[symbol] - target_stoichiometry[symbol]
            else:
                difference = self.get_stoichiometry()[symbol]

            if difference > 0:
                excess_atoms = np.append(excess_atoms, choose_n_atoms_with_symbol(symbol, difference), 0)

        np.random.shuffle(excess_atoms)
        excess_atoms = excess_atoms.astype(np.int)

        for symbol in target_stoichiometry:
            difference = target_stoichiometry[symbol]
            if symbol in self.get_stoichiometry():
                difference = target_stoichiometry[symbol] - self.get_stoichiometry()[symbol]
            self.transform_atoms(excess_atoms[-difference:], [symbol]*difference)
            excess_atoms = excess_atoms[:-difference]
        return






"""
NPlib - A library for numerical processing.

This package provides various numerical processing utilities.
"""

__version__ = "0.1.0"
__author__ = "Riccardo Farris"
__credits__ = 'Reactive Nanoparticles Laboratory'



import numpy as np


class RandomExchangeOperator:
    def __init__(self, p_geometric):
        self.symbol1 = None
        self.symbol2 = None

        self.n_symbol1_atoms = 0
        self.n_symbol2_atoms = 0

        self.n_adsorbates = 0
        self.n_adsorption_sites = 0

        self.max_exchanges = 0
        self.max_adsorbate_exchanges = 0
        self.p_geometric = p_geometric

    def bind_particle(self, particle):
        symbols = sorted(particle.atoms.get_all_symbols())
        self.symbol1 = symbols[0]
        if len(symbols) == 2:
            self.symbol2 = symbols[1]

        symbol1_indices = particle.get_indices_by_symbol(self.symbol1)
        symbol2_indices = particle.get_indices_by_symbol(self.symbol2)

        self.n_symbol1_atoms = len(symbol1_indices)
        self.n_symbol2_atoms = len(symbol2_indices)

        self.max_exchanges = min(self.n_symbol1_atoms, self.n_symbol2_atoms)

    def bind_adsorbates(self, particle, n_adsorbates):
        self.bind_particle(particle)
        particle.construct_adsorption_list()

        if particle.get_number_of_adsorbates() != n_adsorbates:
            particle.random_occupation(n_adsorbates)

        self.n_adsorbates = n_adsorbates
        self.n_adsorption_sites = particle.get_total_number_of_sites()
        self.max_adsorbate_exchanges = n_adsorbates

    def random_exchange(self, particle):
        n_exchanges = min(np.random.geometric(p=self.p_geometric, size=1)[0], self.max_exchanges)
        symbol1_indices = np.random.choice(particle.get_indices_by_symbol(self.symbol1), n_exchanges, replace=False)
        symbol2_indices = np.random.choice(particle.get_indices_by_symbol(self.symbol2), n_exchanges, replace=False)

        particle.swap_symbols(zip(symbol1_indices, symbol2_indices))
        return list(zip(symbol1_indices, symbol2_indices))

    def random_adsorbate_migration(self, particle):

        n_exchanges = min(np.random.geometric(p=self.p_geometric, size=1)[0], self.max_adsorbate_exchanges)
        status0_indices = np.random.choice(particle.get_occupation_status_by_indices(0), n_exchanges, replace=False)
        status1_indices = np.random.choice(particle.get_occupation_status_by_indices(1), n_exchanges, replace=False)

        particle.swap_status(zip(status0_indices, status1_indices))
        return list(zip(status0_indices, status1_indices))

    def coupled_random_exchange(self, particle):
        probability = np.random.randint(2)
        n_exchanges = min(np.random.geometric(p=self.p_geometric, size=1)[0], self.max_exchanges)
        symbol1_indices = np.random.choice(particle.get_indices_by_symbol(self.symbol1), n_exchanges*probability, replace=False)
        symbol2_indices = np.random.choice(particle.get_indices_by_symbol(self.symbol2), n_exchanges*probability, replace=False)

        ordering_exchange = list(zip(symbol1_indices, symbol2_indices))
        particle.swap_symbols(ordering_exchange)

        n_exchanges = min(np.random.geometric(p=self.p_geometric, size=1)[0], self.max_adsorbate_exchanges)
        status0_indices = np.random.choice(particle.get_occupation_status_by_indices(0), n_exchanges*(1-probability), replace=False)
        status1_indices = np.random.choice(particle.get_occupation_status_by_indices(1), n_exchanges*(1-probability), replace=False)

        adsorbate_exchange = list(zip(status0_indices, status1_indices))
        particle.swap_status(adsorbate_exchange)

        return ordering_exchange, adsorbate_exchange
